@@ -36,12 +36,18 @@ void bcachefs_usage(void)
 	     "  recover-super            Attempt to recover overwritten superblock from backups\n"
 	     "  set-fs-option            Set a filesystem option\n"
 	     "  reset-counters           Reset all counters on an unmounted device\n"
+	     "  strip-alloc              Strip alloc info on a filesystem to be used read-only\n"
+	     "\n"
+	     "Commands for managing images:\n"
+	     "  image create             Create a new compact disk image\n"
+	     "  image update             Sync an image with a directory tree\n"
 	     "\n"
 	     "Mount:\n"
 	     "  mount                    Mount a filesystem\n"
 	     "\n"
 	     "Repair:\n"
 	     "  fsck                     Check an existing filesystem for errors\n"
+	     "  recovery-pass            Schedule or deschedule recovery passes\n"
 	     "\n"
 #if 0
 	     "Startup/shutdown, assembly of multi device filesystems:\n"
@@ -102,18 +108,3 @@ void bcachefs_usage(void)
          "  completions              Generate shell completions\n"
 	     "  version                  Display the version of the invoked bcachefs tool\n");
 }
-
-int fs_cmds(int argc, char *argv[])
-{
-	char *cmd = pop_cmd(&argc, argv);
-
-	if (argc < 1)
-		return fs_usage();
-	if (!strcmp(cmd, "usage"))
-		return cmd_fs_usage(argc, argv);
-	if (!strcmp(cmd, "top"))
-		return cmd_fs_top(argc, argv);
-
-	fs_usage();
-	return -EINVAL;
-}
