@@ -12,6 +12,8 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+#include <blkid.h>
+
 #include "libbcachefs/bcachefs.h"
 #include "libbcachefs/bcachefs_ioctl.h"
 #include "libbcachefs/errcode.h"
@@ -47,7 +49,6 @@ static int cmd_device_add(int argc, char *argv[])
 		{ "help",		no_argument,		NULL, 'h' },
 		{ NULL }
 	};
-	struct format_opts format_opts	= format_opts_default();
 	struct dev_opts dev_opts	= dev_opts_default();
 	bool force = false;
 
@@ -88,7 +89,7 @@ static int cmd_device_add(int argc, char *argv[])
 			exit(EXIT_FAILURE);
 			break;
 		}
-}
+	}
 	args_shift(optind);
 
 	char *fs_path = arg_pop();
@@ -104,27 +105,26 @@ static int cmd_device_add(int argc, char *argv[])
 
 	struct bchfs_handle fs = bcache_fs_open(fs_path);
 
-	int ret = open_for_format(&dev_opts, force);
+	int ret = open_for_format(&dev_opts, 0, force) ?:
+		bch2_format_for_device_add(&dev_opts,
+			read_file_u64(fs.sysfs_fd, "options/block_size"),
+			read_file_u64(fs.sysfs_fd, "options/btree_node_size"));
 	if (ret)
 		die("Error opening %s: %s", dev_opts.path, strerror(-ret));
 
-	struct bch_opt_strs fs_opt_strs;
-	memset(&fs_opt_strs, 0, sizeof(fs_opt_strs));
-
-	struct bch_opts fs_opts = bch2_parse_opts(fs_opt_strs);
-
-	opt_set(fs_opts, block_size,
-		read_file_u64(fs.sysfs_fd, "options/block_size"));
-	opt_set(fs_opts, btree_node_size,
-		read_file_u64(fs.sysfs_fd, "options/btree_node_size"));
-
-	dev_opts_list devs = {};
-	darray_push(&devs, dev_opts);
-
-	struct bch_sb *sb = bch2_format(fs_opt_strs, fs_opts, format_opts, devs);
-	darray_exit(&devs);
-	free(sb);
 	bchu_disk_add(fs, dev_opts.path);
+
+	/* A whole bunch of nonsense to get blkid to update its cache, so
+	 * mount can find the new device by UUID:
+	 */
+	blkid_cache cache = NULL;
+	if (!blkid_get_cache(&cache, NULL)) {
+		blkid_dev dev = blkid_get_dev(cache, dev_opts.path, BLKID_DEV_VERIFY);
+		if (dev)
+			blkid_verify(cache, dev);
+		blkid_put_cache(cache);
+	}
+
 	return 0;
 }
 
@@ -212,9 +212,13 @@ static void device_online_usage(void)
 
 static int cmd_device_online(int argc, char *argv[])
 {
+	static const struct option longopts[] = {
+		{ "help",		0, NULL, 'h' },
+		{ NULL }
+	};
 	int opt;
 
-	while ((opt = getopt(argc, argv, "h")) != -1)
+	while ((opt = getopt_long(argc, argv, "h", longopts, NULL)) != -1)
 		switch (opt) {
 		case 'h':
 			device_online_usage();
@@ -286,6 +290,7 @@ static void device_evacuate_usage(void)
 	     "Usage: bcachefs device evacuate [OPTION]... device\n"
 	     "\n"
 	     "Options:\n"
+	     "  -f, --force		    Force if data redundancy will be degraded\n"
 	     "  -h, --help                  Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
@@ -293,10 +298,18 @@ static void device_evacuate_usage(void)
 
 static int cmd_device_evacuate(int argc, char *argv[])
 {
-	int opt;
+	static const struct option longopts[] = {
+		{ "force",			no_argument,	NULL, 'f' },
+		{ "help",			no_argument,	NULL, 'h' },
+		{ NULL }
+	};
+	int opt, flags = 0;
 
-	while ((opt = getopt(argc, argv, "h")) != -1)
+	while ((opt = getopt_long(argc, argv, "fh", longopts, NULL)) != -1)
 		switch (opt) {
+		case 'f':
+			flags |= BCH_FORCE_IF_DEGRADED;
+			break;
 		case 'h':
 			device_evacuate_usage();
 			exit(EXIT_SUCCESS);
@@ -317,7 +330,7 @@ static int cmd_device_evacuate(int argc, char *argv[])
 
 	if (u->state == BCH_MEMBER_STATE_rw) {
 		printf("Setting %s readonly\n", dev_path);
-		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, 0);
+		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, flags);
 	}
 
 	free(u);
@@ -341,8 +354,8 @@ static void device_set_state_usage(void)
 	     "<path>: path to mounted filesystem, optional unless specifying device by id\n"
 	     "\n"
 	     "Options:\n"
-	     "  -f, --force		    Force, if data redundancy will be degraded\n"
-	     "      --force-if-data-lost    Force, if data will be lost\n"
+	     "  -f, --force		    Force if data redundancy will be degraded\n"
+	     "      --force-if-data-lost    Force if data will be lost\n"
 	     "  -o, --offline               Set state of an offline device\n"
 	     "  -h, --help                  display this help and exit\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
@@ -352,10 +365,10 @@ static void device_set_state_usage(void)
 static int cmd_device_set_state(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "force",			0, NULL, 'f' },
-		{ "force-if-data-lost",		0, NULL, 'F' },
-		{ "offline",			0, NULL, 'o' },
-		{ "help",			0, NULL, 'h' },
+		{ "force",			no_argument,	NULL, 'f' },
+		{ "force-if-data-lost",		no_argument,	NULL, 'F' },
+		{ "offline",			no_argument,	NULL, 'o' },
+		{ "help",			no_argument,	NULL, 'h' },
 		{ NULL }
 	};
 	struct bchfs_handle fs;
@@ -505,7 +518,7 @@ static int cmd_device_resize(int argc, char *argv[])
 
 		struct bch_member m = bch2_sb_member_get(sb, idx);
 
-		u64 nbuckets = size / le16_to_cpu(m.bucket_size);
+		u64 nbuckets = size / BCH_MEMBER_BUCKET_SIZE(&m);
 
 		if (nbuckets < le64_to_cpu(m.nbuckets))
 			die("Shrinking not supported yet");
@@ -515,20 +528,24 @@ static int cmd_device_resize(int argc, char *argv[])
 	} else {
 		printf("Doing offline resize of %s\n", dev);
 
-		struct bch_fs *c = bch2_fs_open(&dev, 1, bch2_opts_empty());
+		darray_const_str devs = {};
+		darray_push(&devs, dev);
+
+		struct bch_opts opts = bch2_opts_empty();
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
 		if (IS_ERR(c))
 			die("error opening %s: %s", dev, bch2_err_str(PTR_ERR(c)));
 
 		struct bch_dev *resize = NULL;
 
-		for_each_online_member(c, ca) {
+		for_each_online_member(c, ca, 0) {
 			if (resize)
 				die("confused: more than one online device?");
 			resize = ca;
-			percpu_ref_get(&resize->io_ref[READ]);
+			enumerated_ref_get(&resize->io_ref[READ], 0);
 		}
 
-		u64 nbuckets = size / le16_to_cpu(resize->mi.bucket_size);
+		u64 nbuckets = size / resize->mi.bucket_size;
 
 		if (nbuckets < le64_to_cpu(resize->mi.nbuckets))
 			die("Shrinking not supported yet");
@@ -538,7 +555,7 @@ static int cmd_device_resize(int argc, char *argv[])
 		if (ret)
 			fprintf(stderr, "resize error: %s\n", bch2_err_str(ret));
 
-		percpu_ref_put(&resize->io_ref[READ]);
+		enumerated_ref_put(&resize->io_ref[READ], 0);
 		bch2_fs_stop(c);
 	}
 	return 0;
@@ -605,24 +622,28 @@ static int cmd_device_resize_journal(int argc, char *argv[])
 
 		struct bch_member m = bch2_sb_member_get(sb, idx);
 
-		u64 nbuckets = size / le16_to_cpu(m.bucket_size);
+		u64 nbuckets = size / BCH_MEMBER_BUCKET_SIZE(&m);
 
 		printf("resizing journal on %s to %llu buckets\n", dev, nbuckets);
 		bchu_disk_resize_journal(fs, idx, nbuckets);
 	} else {
 		printf("%s is offline - starting:\n", dev);
 
-		struct bch_fs *c = bch2_fs_open(&dev, 1, bch2_opts_empty());
+		darray_const_str devs = {};
+		darray_push(&devs, dev);
+
+		struct bch_opts opts = bch2_opts_empty();
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
 		if (IS_ERR(c))
 			die("error opening %s: %s", dev, bch2_err_str(PTR_ERR(c)));
 
 		struct bch_dev *resize = NULL;
 
-		for_each_online_member(c, ca) {
+		for_each_online_member(c, ca, 0) {
 			if (resize)
 				die("confused: more than one online device?");
 			resize = ca;
-			percpu_ref_get(&resize->io_ref[READ]);
+			enumerated_ref_get(&resize->io_ref[READ], 0);
 		}
 
 		u64 nbuckets = size / le16_to_cpu(resize->mi.bucket_size);
@@ -632,7 +653,7 @@ static int cmd_device_resize_journal(int argc, char *argv[])
 		if (ret)
 			fprintf(stderr, "resize error: %s\n", bch2_err_str(ret));
 
-		percpu_ref_put(&resize->io_ref[READ]);
+		enumerated_ref_put(&resize->io_ref[READ], 0);
 		bch2_fs_stop(c);
 	}
 	return 0;
