@@ -6,27 +6,11 @@
 #include "cmds.h"
 #include "libbcachefs/error.h"
 #include "libbcachefs.h"
+#include "libbcachefs/recovery_passes.h"
 #include "libbcachefs/super.h"
 #include "libbcachefs/super-io.h"
 #include "tools-util.h"
 
-static void fsck_usage(void)
-{
-	puts("bcachefs fsck - filesystem check and repair\n"
-	     "Usage: bcachefs fsck [OPTION]... <devices>\n"
-	     "\n"
-	     "Options:\n"
-	     "  -p                      Automatic repair (no questions)\n"
-	     "  -n                      Don't repair, only check for errors\n"
-	     "  -y                      Assume \"yes\" to all questions\n"
-	     "  -f                      Force checking even if filesystem is marked clean\n"
-	     "  -r, --ratelimit_errors  Don't display more than 10 errors of a given type\n"
-	     "  -k, --kernel            Use the in-kernel fsck implementation\n"
-	     "  -v                      Be verbose\n"
-	     "  -h, --help              Display this help and exit\n"
-	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
-}
-
 static void setnonblocking(int fd)
 {
 	int flags = fcntl(fd, F_GETFL);
@@ -47,6 +31,24 @@ static int do_splice(int rfd, int wfd)
 		return 1;
 	do {
 		ssize_t w = write(wfd, b, r);
+
+		/*
+		 * Ugly, but we have no way of doing nonblocking reads and
+		 * blocking writes.
+		 *
+		 * Yes, this means that if one thread has stopped reading (or
+		 * isn't keeping up) we block traffic on the other direction of
+		 * the pipe. No, I don't care.
+		 */
+		if (w < 0 && errno == EAGAIN) {
+			fd_set fds;
+			FD_ZERO(&fds);
+			FD_SET(wfd, &fds);
+			if (select(wfd + 1, NULL, &fds, NULL, NULL) < 0)
+				die("select error: %m");
+			continue;
+		}
+
 		if (w < 0)
 			die("%s: write error: %m", __func__);
 		r -= w;
@@ -86,15 +88,12 @@ static int splice_fd_to_stdinout(int fd)
 			stdin_closed = true;
 	}
 
-	xclose(fd);
-	return 0;
+	/* the return code from fsck itself is returned via close() */
+	return close(fd);
 }
 
-static int fsck_online(const char *dev_path, const char *opt_str)
+static int fsck_online(struct bchfs_handle fs, const char *opt_str)
 {
-	int dev_idx;
-	struct bchfs_handle fs = bchu_fs_open_by_dev(dev_path, &dev_idx);
-
 	struct bch_ioctl_fsck_online fsck = {
 		.opts = (unsigned long) opt_str
 	};
@@ -113,10 +112,8 @@ static void append_opt(struct printbuf *out, const char *opt)
 	prt_str(out, opt);
 }
 
-static bool should_use_kernel_fsck(darray_str devs)
+static bool should_use_kernel_fsck(darray_const_str devs)
 {
-	system("modprobe bcachefs");
-
 	unsigned kernel_version = bcachefs_kernel_version();
 
 	if (!kernel_version)
@@ -131,7 +128,7 @@ static bool should_use_kernel_fsck(darray_str devs)
 	opt_set(opts, nochanges, true);
 	opt_set(opts, read_only, true);
 
-	struct bch_fs *c = bch2_fs_open(devs.data, devs.nr, opts);
+	struct bch_fs *c = bch2_fs_open(&devs, &opts);
 	if (IS_ERR(c))
 		return false;
 
@@ -199,6 +196,23 @@ static char *loopdev_alloc(const char *path)
 	return line;
 }
 
+static void fsck_usage(void)
+{
+	puts("bcachefs fsck - filesystem check and repair\n"
+	     "Usage: bcachefs fsck [OPTION]... <devices>\n"
+	     "\n"
+	     "Options:\n"
+	     "  -p                      Automatic repair (no questions)\n"
+	     "  -n                      Don't repair, only check for errors\n"
+	     "  -y                      Assume \"yes\" to all questions\n"
+	     "  -f                      Force checking even if filesystem is marked clean\n"
+	     "  -r, --ratelimit_errors  Don't display more than 10 errors of a given type\n"
+	     "  -k, --kernel            Use the in-kernel fsck implementation\n"
+	     "  -v                      Be verbose\n"
+	     "  -h, --help              Display this help and exit\n"
+	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+}
+
 int cmd_fsck(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
@@ -225,10 +239,10 @@ int cmd_fsck(int argc, char *argv[])
 				  longopts, NULL)) != -1)
 		switch (opt) {
 		case 'a':
+		case 'p':
 			/* "automatic" run, called by the system, for us to do checks as needed.
 			 *  we don't need checks here: */
 			exit(EXIT_SUCCESS);
-		case 'p':
 		case 'y':
 			append_opt(&opts_str, "fix_errors=yes");
 			break;
@@ -265,14 +279,28 @@ int cmd_fsck(int argc, char *argv[])
 		exit(8);
 	}
 
-	darray_str devs = get_or_split_cmdline_devs(argc, argv);
+	darray_const_str devs = get_or_split_cmdline_devs(argc, argv);
+
+	if (devs.nr == 1 &&
+	    S_ISDIR(xstat(devs.data[0]).st_mode)) {
+		printf("Running fsck online\n");
+
+		struct bchfs_handle fs = bcache_fs_open(devs.data[0]);
+		return fsck_online(fs, opts_str.buf);
+	}
 
 	darray_for_each(devs, i)
 		if (dev_mounted(*i)) {
 			printf("Running fsck online\n");
-			return fsck_online(*i, opts_str.buf);
+
+			int dev_idx;
+			struct bchfs_handle fs = bchu_fs_open_by_dev(*i, &dev_idx);
+			return fsck_online(fs, opts_str.buf);
 		}
 
+	if (kernel)
+		system("modprobe bcachefs");
+
 	int kernel_probed = kernel;
 	if (kernel_probed < 0)
 		kernel_probed = should_use_kernel_fsck(devs);
@@ -324,7 +352,7 @@ userland_fsck:
 		if (ret)
 			return ret;
 
-		struct bch_fs *c = bch2_fs_open(devs.data, devs.nr, opts);
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
 		if (IS_ERR(c))
 			exit(8);
 
@@ -343,3 +371,94 @@ userland_fsck:
 	printbuf_exit(&opts_str);
 	return ret;
 }
+
+static void recovery_pass_usage(void)
+{
+	puts("bcachefs recovery-pass - list and manage scheduled recovery passes\n"
+	     "Usage: bcachefs recovery-pass [OPTION]... <devices>\n"
+	     "\n"
+	     "Currently only supports unmounted/offline filesystems\n"
+	     "\n"
+	     "Options:\n"
+	     "  -s, --set               Schedule a recovery pass in the superblock\n"
+	     "  -u, --unset             Deschedule a recovery pass\n"
+	     "  -h, --help              Display this help and exit\n"
+	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+}
+
+int cmd_recovery_pass(int argc, char *argv[])
+{
+	static const struct option longopts[] = {
+		{ "set",		required_argument,	NULL, 's' },
+		{ "unset",		required_argument,	NULL, 'u' },
+		{ "help",		no_argument,		NULL, 'h' },
+		{ NULL }
+	};
+	u64 passes_to_set = 0, passes_to_unset = 0;
+	int opt;
+
+	while ((opt = getopt_long(argc, argv, "s:u:h", longopts, NULL)) != -1)
+		switch (opt) {
+		case 's':
+			passes_to_set |= read_flag_list_or_die(optarg,
+						bch2_recovery_passes,
+						"recovery pass");
+			break;
+		case 'u':
+			passes_to_unset |= read_flag_list_or_die(optarg,
+						bch2_recovery_passes,
+						"recovery pass");
+			break;
+		case 'h':
+			recovery_pass_usage();
+			exit(EXIT_SUCCESS);
+		}
+	args_shift(optind);
+
+	passes_to_set	= bch2_recovery_passes_to_stable(passes_to_set);
+	passes_to_unset	= bch2_recovery_passes_to_stable(passes_to_unset);
+
+	darray_const_str devs = get_or_split_cmdline_devs(argc, argv);
+
+	struct bch_opts opts = bch2_opts_empty();
+	opt_set(opts, nostart, true);
+
+	struct bch_fs *c = bch2_fs_open(&devs, &opts);
+	int ret = PTR_ERR_OR_ZERO(c);
+	if (ret)
+		die("Error opening filesystem: %s", bch2_err_str(ret));
+
+	scoped_guard(mutex, &c->sb_lock) {
+		struct bch_sb_field_ext *ext =
+			bch2_sb_field_get_minsize(&c->disk_sb, ext,
+				sizeof(struct bch_sb_field_ext) / sizeof(u64));
+		if (!ext) {
+			fprintf(stderr, "Error getting sb_field_ext\n");
+			goto err;
+		}
+
+		u64 scheduled = le64_to_cpu(ext->recovery_passes_required[0]);
+
+		if (passes_to_set || passes_to_unset) {
+			ext->recovery_passes_required[0] &= ~cpu_to_le64(passes_to_unset);
+			ext->recovery_passes_required[0] |=  cpu_to_le64(passes_to_set);
+
+			scheduled = le64_to_cpu(ext->recovery_passes_required[0]);
+
+			bch2_write_super(c);
+		}
+
+		CLASS(printbuf, buf)();
+		prt_str(&buf, "Scheduled recovery passes: ");
+		if (scheduled)
+			prt_bitflags(&buf, bch2_recovery_passes,
+				     bch2_recovery_passes_from_stable(scheduled));
+		else
+			prt_str(&buf, "(none)");
+		printf("%s\n", buf.buf);
+	}
+err:
+	bch2_fs_stop(c);
+	return ret;
+}
+
