@@ -1,5 +1,6 @@
 #include <errno.h>
 #include <fcntl.h>
+#include <getopt.h>
 #include <unistd.h>
 #include <uuid/uuid.h>
 
@@ -15,16 +16,23 @@ static void unlock_usage(void)
 	     "Usage: bcachefs unlock [OPTION] device\n"
 	     "\n"
 	     "Options:\n"
-	     "  -c                     Check if a device is encrypted\n"
-	     "  -k (session|user|user_session)\n"
+	     "  -c, --check            Check if a device is encrypted\n"
+	     "  -k, --keyring (session|user|user_session)\n"
 	     "                         Keyring to add to (default: user)\n"
-	     "  -f                     Passphrase file to read from (disables passphrase prompt)\n"
-	     "  -h                     Display this help and exit\n"
+	     "  -f, --file             Passphrase file to read from (disables passphrase prompt)\n"
+	     "  -h, --help             Display this help and exit\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
 int cmd_unlock(int argc, char *argv[])
 {
+	static const struct option longopts[] = {
+		{ "check",		no_argument,		NULL, 'c' },
+		{ "keyring",		required_argument,	NULL, 'k' },
+		{ "file",		required_argument,	NULL, 'f' },
+		{ "help",		no_argument,		NULL, 'h' },
+		{ NULL }
+	};
 	const char *keyring = "user";
 	bool check = false;
 	const char *passphrase_file_path = NULL;
@@ -32,7 +40,7 @@ int cmd_unlock(int argc, char *argv[])
 
 	int opt;
 
-	while ((opt = getopt(argc, argv, "cf:k:h")) != -1)
+	while ((opt = getopt_long(argc, argv, "cf:k:h", longopts, NULL)) != -1)
 		switch (opt) {
 		case 'c':
 			check = true;
@@ -87,12 +95,13 @@ int cmd_unlock(int argc, char *argv[])
 
 int cmd_set_passphrase(int argc, char *argv[])
 {
-	struct bch_opts opts = bch2_opts_empty();
-	struct bch_fs *c;
-
-	if (argc < 2)
+	args_shift(1);
+	if (!argc)
 		die("Please supply one or more devices");
 
+	darray_const_str devs = get_or_split_cmdline_devs(argc, argv);
+
+	struct bch_opts opts = bch2_opts_empty();
 	opt_set(opts, nostart, true);
 
 	/*
@@ -100,7 +109,7 @@ int cmd_set_passphrase(int argc, char *argv[])
 	 * to make sure we're opening and updating every component device:
 	 */
 
-	c = bch2_fs_open(argv + 1, argc - 1, opts);
+	struct bch_fs *c = bch2_fs_open(&devs, &opts);
 	if (IS_ERR(c))
 		die("Error opening %s: %s", argv[1], bch2_err_str(PTR_ERR(c)));
 
@@ -126,14 +135,16 @@ int cmd_set_passphrase(int argc, char *argv[])
 
 int cmd_remove_passphrase(int argc, char *argv[])
 {
-	struct bch_opts opts = bch2_opts_empty();
-	struct bch_fs *c;
-
-	if (argc < 2)
+	args_shift(1);
+	if (!argc)
 		die("Please supply one or more devices");
 
+	darray_const_str devs = get_or_split_cmdline_devs(argc, argv);
+
+	struct bch_opts opts = bch2_opts_empty();
 	opt_set(opts, nostart, true);
-	c = bch2_fs_open(argv + 1, argc - 1, opts);
+
+	struct bch_fs *c = bch2_fs_open(&devs, &opts);
 	if (IS_ERR(c))
 		die("Error opening %s: %s", argv[1], bch2_err_str(PTR_ERR(c)));
 
