@@ -244,7 +244,6 @@ static int migrate_fs(const char		*fs_path,
 	find_superblock_space(extents, format_opts, dev);
 
 	struct bch_sb *sb = bch2_format(fs_opt_strs, fs_opts, format_opts, devs);
-	darray_exit(&devs);
 
 	u64 sb_offset = le64_to_cpu(sb->layout.sb_offset[0]);
 
@@ -253,14 +252,15 @@ static int migrate_fs(const char		*fs_path,
 
 	free(sb);
 
-	char *path[1] = { dev->path };
+	darray_const_str dev_paths = {};
+	darray_push(&dev_paths, dev->path);
 
 	struct bch_opts opts = bch2_opts_empty();
 	opt_set(opts, sb,	sb_offset);
 	opt_set(opts, nostart,	true);
 	opt_set(opts, noexcl,	true);
 
-	struct bch_fs *c = bch2_fs_open(path, 1, opts);
+	struct bch_fs *c = bch2_fs_open(&dev_paths, &opts);
 	if (IS_ERR(c))
 		die("Error opening new filesystem: %s", bch2_err_str(PTR_ERR(c)));
 
@@ -279,23 +279,25 @@ static int migrate_fs(const char		*fs_path,
 		.dev		= stat.st_dev,
 		.extents	= extents,
 		.type		= BCH_MIGRATE_migrate,
+		.reserve_start	= roundup((format_opts.superblock_size * 2 + 8) << 9,
+					  bucket_bytes(c->devs[0])),
 	};
 
+	BUG_ON(!s.reserve_start);
 
-	u64 reserve_start = roundup((format_opts.superblock_size * 2 + 8) << 9,
-				    bucket_bytes(c->devs[0]));
-	BUG_ON(!reserve_start);
-
-	copy_fs(c, fs_fd, fs_path, &s, reserve_start);
+	ret = copy_fs(c, &s, fs_fd, fs_path);
 
 	bch2_fs_stop(c);
 
+	if (ret)
+		return ret;
+
 	printf("Migrate complete, running fsck:\n");
 	opt_set(opts, nostart,	false);
 	opt_set(opts, nochanges, true);
 	opt_set(opts, read_only, true);
 
-	c = bch2_fs_open(path, 1, opts);
+	c = bch2_fs_open(&dev_paths, &opts);
 	if (IS_ERR(c))
 		die("Error opening new filesystem: %s", bch2_err_str(PTR_ERR(c)));
 
@@ -315,6 +317,8 @@ static int migrate_fs(const char		*fs_path,
 	       "no longer needed (and should be deleted prior to running\n"
 	       "bcachefs migrate-superblock)\n",
 	       sb_offset, dev->path, dev->path, sb_offset);
+
+	darray_exit(&devs);
 	return 0;
 }
 
@@ -369,22 +373,28 @@ static void migrate_superblock_usage(void)
 	     "Usage: bcachefs migrate-superblock [OPTION]...\n"
 	     "\n"
 	     "Options:\n"
-	     "  -d device     Device to create superblock for\n"
-	     "  -o offset     Offset of existing superblock\n"
-	     "  -h            Display this help and exit\n"
+	     "  -d, --dev    device     Device to create superblock for\n"
+	     "  -o, --offset offset     Offset of existing superblock\n"
+	     "  -h, --help              Display this help and exit\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
 int cmd_migrate_superblock(int argc, char *argv[])
 {
-	char *dev = NULL;
+	static const struct option longopts[] = {
+		{ "dev",		required_argument,	NULL, 'd' },
+		{ "offset",		required_argument,	NULL, 'o' },
+		{ "help",		no_argument,		NULL, 'h' },
+		{ NULL }
+	};
+	darray_const_str devs = {};
 	u64 sb_offset = 0;
 	int opt, ret;
 
-	while ((opt = getopt(argc, argv, "d:o:h")) != -1)
+	while ((opt = getopt_long(argc, argv, "d:o:h", longopts, NULL)) != -1)
 		switch (opt) {
 			case 'd':
-				dev = optarg;
+				darray_push(&devs, optarg);
 				break;
 			case 'o':
 				ret = kstrtou64(optarg, 10, &sb_offset);
@@ -396,13 +406,13 @@ int cmd_migrate_superblock(int argc, char *argv[])
 				exit(EXIT_SUCCESS);
 		}
 
-	if (!dev)
+	if (!devs.nr)
 		die("Please specify a device");
 
 	if (!sb_offset)
 		die("Please specify offset of existing superblock");
 
-	int fd = xopen(dev, O_RDWR);
+	int fd = xopen(devs.data[0], O_RDWR);
 	struct bch_sb *sb = __bch2_super_read(fd, sb_offset);
 	unsigned sb_size = 1U << sb->layout.sb_max_size_bits;
 
@@ -435,7 +445,7 @@ int cmd_migrate_superblock(int argc, char *argv[])
 	opt_set(opts, nostart,	true);
 	opt_set(opts, sb,	sb_offset);
 
-	struct bch_fs *c = bch2_fs_open(&dev, 1, opts);
+	struct bch_fs *c = bch2_fs_open(&devs, &opts);
 	ret =   PTR_ERR_OR_ZERO(c) ?:
 		bch2_buckets_nouse_alloc(c);
 	if (ret)
@@ -461,7 +471,7 @@ int cmd_migrate_superblock(int argc, char *argv[])
 	 * inconsequential:
 	 */
 
-	c = bch2_fs_open(&dev, 1, opts);
+	c = bch2_fs_open(&devs, &opts);
 	ret =   PTR_ERR_OR_ZERO(c);
 	if (ret)
 		die("error opening filesystem: %s", bch2_err_str(ret));
