@@ -13,8 +13,11 @@ int cmd_format(int argc, char *argv[]);
 int cmd_show_super(int argc, char *argv[]);
 int cmd_recover_super(int argc, char *argv[]);
 int cmd_reset_counters(int argc, char *argv[]);
+int cmd_strip_alloc(int argc, char *argv[]);
 int cmd_set_option(int argc, char *argv[]);
 
+int image_cmds(int argc, char *argv[]);
+
 int fs_usage(void);
 int cmd_fs_usage(int argc, char *argv[]);
 int cmd_fs_top(int argc, char *argv[]);
@@ -28,6 +31,7 @@ int cmd_set_passphrase(int argc, char *argv[]);
 int cmd_remove_passphrase(int argc, char *argv[]);
 
 int cmd_fsck(int argc, char *argv[]);
+int cmd_recovery_pass(int argc, char *argv[]);
 
 int cmd_dump(int argc, char *argv[]);
 int cmd_list_journal(int argc, char *argv[]);
