@@ -36,8 +36,10 @@
 
 void bch2_sb_layout_init(struct bch_sb_layout *l,
 			 unsigned block_size,
+			 unsigned bucket_size,
 			 unsigned sb_size,
-			 u64 sb_start, u64 sb_end)
+			 u64 sb_start, u64 sb_end,
+			 bool no_sb_at_end)
 {
 	u64 sb_pos = sb_start;
 	unsigned i;
@@ -61,6 +63,20 @@ void bch2_sb_layout_init(struct bch_sb_layout *l,
 	if (sb_pos > sb_end)
 		die("insufficient space for superblocks: start %llu end %llu > %llu size %u",
 		    sb_start, sb_pos, sb_end, sb_size);
+
+	/*
+	 * Also create a backup superblock at the end of the disk:
+	 *
+	 * If we're not creating a superblock at the default offset, it
+	 * means we're being run from the migrate tool and we could be
+	 * overwriting existing data if we write to the end of the disk:
+	 */
+	if (sb_start == BCH_SB_SECTOR && !no_sb_at_end) {
+		u64 backup_sb = sb_end - (1 << l->sb_max_size_bits);
+
+		backup_sb = rounddown(backup_sb, bucket_size >> 9);
+		l->sb_offset[l->nr_superblocks++] = cpu_to_le64(backup_sb);
+	}
 }
 
 static u64 dev_max_bucket_size(u64 dev_size)
@@ -222,7 +238,7 @@ struct bch_sb *bch2_format(struct bch_opt_strs	fs_opt_strs,
 	sb.sb->nr_devices	= devs.nr;
 	SET_BCH_SB_VERSION_INCOMPAT_ALLOWED(sb.sb, opts.version);
 
-	if (opts.version == bcachefs_metadata_version_current)
+	if (opts.version > bcachefs_metadata_version_disk_accounting_big_endian)
 		sb.sb->features[0] |= cpu_to_le64(BCH_SB_FEATURES_ALL);
 
 	uuid_generate(sb.sb->uuid.b);
@@ -304,24 +320,12 @@ struct bch_sb *bch2_format(struct bch_opt_strs	fs_opt_strs,
 			i->sb_end	= size_sectors;
 		}
 
-		bch2_sb_layout_init(&sb.sb->layout, fs_opts.block_size,
+		bch2_sb_layout_init(&sb.sb->layout,
+				    fs_opts.block_size,
+				    i->opts.bucket_size,
 				    opts.superblock_size,
-				    i->sb_offset, i->sb_end);
-
-		/*
-		 * Also create a backup superblock at the end of the disk:
-		 *
-		 * If we're not creating a superblock at the default offset, it
-		 * means we're being run from the migrate tool and we could be
-		 * overwriting existing data if we write to the end of the disk:
-		 */
-		if (i->sb_offset == BCH_SB_SECTOR) {
-			struct bch_sb_layout *l = &sb.sb->layout;
-			u64 backup_sb = size_sectors - (1 << l->sb_max_size_bits);
-
-			backup_sb = rounddown(backup_sb, i->opts.bucket_size >> 9);
-			l->sb_offset[l->nr_superblocks++] = cpu_to_le64(backup_sb);
-		}
+				    i->sb_offset, i->sb_end,
+				    opts.no_sb_at_end);
 
 		if (i->sb_offset == BCH_SB_SECTOR) {
 			/* Zero start of disk */
@@ -338,6 +342,27 @@ struct bch_sb *bch2_format(struct bch_opt_strs	fs_opt_strs,
 	return sb.sb;
 }
 
+int bch2_format_for_device_add(struct dev_opts *dev,
+			       unsigned block_size, unsigned btree_node_size)
+{
+	struct bch_opt_strs fs_opt_strs;
+	memset(&fs_opt_strs, 0, sizeof(fs_opt_strs));
+
+	struct bch_opts fs_opts = bch2_parse_opts(fs_opt_strs);
+	opt_set(fs_opts, block_size,		block_size);
+	opt_set(fs_opts, btree_node_size,	btree_node_size);
+
+	dev_opts_list devs = {};
+	darray_push(&devs, *dev);
+
+	struct format_opts format_opts	= format_opts_default();
+	struct bch_sb *sb = bch2_format(fs_opt_strs, fs_opts, format_opts, devs);
+	darray_exit(&devs);
+	free(sb);
+
+	return 0;
+}
+
 void bch2_super_write(int fd, struct bch_sb *sb)
 {
 	struct nonce nonce = { 0 };
@@ -712,6 +737,12 @@ noopt:
 	return NULL;
 }
 
+void bch_remove_arg_from_argv(int *argc, char *argv[], int index)
+{
+	memmove(&argv[index], &argv[index + 1], (*argc - index) * sizeof(char*));
+	(*argc)--;
+}
+
 struct bch_opt_strs bch2_cmdline_opts_get(int *argc, char *argv[],
 					  unsigned opt_types)
 {
