@@ -31,6 +31,7 @@ void bch2_opt_strs_free(struct bch_opt_strs *);
 
 const struct bch_option *bch2_cmdline_opt_parse(int argc, char *argv[],
 						unsigned opt_types);
+void bch_remove_arg_from_argv(int *argc, char *argv[], int index);
 struct bch_opt_strs bch2_cmdline_opts_get(int *, char *[], unsigned);
 struct bch_opts bch2_parse_opts(struct bch_opt_strs);
 void bch2_opts_usage(unsigned);
@@ -41,8 +42,10 @@ struct format_opts {
 	unsigned	version;
 	unsigned	superblock_size;
 	bool		encrypted;
+	char		*passphrase_file;
 	char		*passphrase;
 	char		*source;
+	bool		no_sb_at_end;
 };
 
 static inline unsigned bcachefs_kernel_version(void)
@@ -58,7 +61,7 @@ static inline struct format_opts format_opts_default()
 	 * Ensure bcachefs module is loaded so we know the supported on disk
 	 * format version:
 	 */
-	system("modprobe bcachefs > /dev/null 2>&1");
+	(void)!system("modprobe bcachefs > /dev/null 2>&1");
 
 	return (struct format_opts) {
 		.version		= bcachefs_kernel_version() ?:
@@ -70,7 +73,7 @@ static inline struct format_opts format_opts_default()
 struct dev_opts {
 	struct file	*file;
 	struct block_device *bdev;
-	char		*path;
+	const char	*path;
 
 	u64		sb_offset;
 	u64		sb_end;
@@ -91,7 +94,7 @@ static inline struct dev_opts dev_opts_default()
 }
 
 void bch2_sb_layout_init(struct bch_sb_layout *,
-			 unsigned, unsigned, u64, u64);
+			 unsigned, unsigned, unsigned, u64, u64, bool);
 
 u64 bch2_pick_bucket_size(struct bch_opts, dev_opts_list);
 void bch2_check_bucket_size(struct bch_opts, struct dev_opts *);
@@ -101,6 +104,9 @@ struct bch_sb *bch2_format(struct bch_opt_strs,
 			   struct format_opts,
 			   dev_opts_list devs);
 
+int bch2_format_for_device_add(struct dev_opts *,
+			       unsigned, unsigned);
+
 void bch2_super_write(int, struct bch_sb *);
 struct bch_sb *__bch2_super_read(int, u64);
 
@@ -124,7 +130,7 @@ struct bchfs_handle bchu_fs_open_by_dev(const char *, int *);
 
 int bchu_dev_path_to_idx(struct bchfs_handle, const char *);
 
-static inline void bchu_disk_add(struct bchfs_handle fs, char *dev)
+static inline void bchu_disk_add(struct bchfs_handle fs, const char *dev)
 {
 	struct bch_ioctl_disk i = { .dev = (unsigned long) dev, };
 
