@@ -376,7 +376,24 @@ pub struct Cli {
     verbose: u8,
 }
 
+fn require_bcachefs_module() {
+    let path = Path::new("/sys/module/bcachefs");
+
+    if !path.exists() {
+        let _ = std::process::Command::new("modprobe")
+            .arg("bcachefs")
+            .status();
+
+        if !path.exists() {
+            error!("bcachefs module not loaded?");
+            std::process::exit(1);
+        }
+    }
+}
+
 pub fn mount(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> std::process::ExitCode {
+    require_bcachefs_module();
+
     // If the bcachefs tool is being called as "bcachefs mount dev ..." (as opposed to via a
     // symlink like "/usr/sbin/mount.bcachefs dev ...", then we need to pop the 0th argument
     // ("bcachefs") since the CLI parser here expects the device at position 1.
